from typing import Union

from rest_framework.serializers import ListSerializer

from rest_serializer_to_typescript import Transpiler
from .type_hinting_mapping import type_hinting_mapping

INDENT = 4


class SerializerMetadataTypeGenerator:
    keep_value_fields = ["type", "field_name"]

    def __init__(self, serializer, metadata, indent=1):
        self.indent = indent
        self.metadata = metadata
        self.serializer = serializer

    def generate_fields_types(self, name, data):
        fields = []

        for field_name, value in data.items():
            if field_name in self.keep_value_fields:
                ts_value = f'"{value}"'
            else:
                ts_value = type_hinting_mapping.get(type(value).__name__, "unknown")

            if field_name == "child":
                sub_serializer = self.serializer.fields[name]

                ts_value = SerializerMetadataTypeGenerator(
                    serializer=self.serializer.fields[name].child,
                    metadata=self.metadata
                ).get_namespace_name()

                if isinstance(sub_serializer, ListSerializer) and sub_serializer.many:
                    ts_value += "[]"

            fields.append(f"{field_name}: {ts_value};")

        return fields

    def generate_field_interface(self, name, data: dict[str, Union[bool, str]]):
        joined_fields = f"\n{' ' * (INDENT * (1 + self.indent))}".join(self.generate_fields_types(name, data))

        pre = ""
        if "child" in data:
            sub_serializer = self.serializer.fields[name].child

            pre = SerializerMetadataTypeGenerator(
                serializer=sub_serializer,
                metadata=self.metadata,
                indent=self.indent + 1
            ).generate_schema()
            pre += f"\n\n{' ' * (INDENT * self.indent)}"

        return f"""{pre}export interface {name} {{
{' ' * (INDENT * (1 + self.indent))}{joined_fields}
{' ' * (INDENT * self.indent)}}}"""

    def generate_schema(self):
        if self.metadata is None:
            raise Exception("Please provide rest framework metadata class")

        fields = self.metadata.get_serializer_info(self.serializer)

        fields_interfaces = f"\n\n{' ' * (INDENT * self.indent)}".join([
            self.generate_field_interface(field_name, field_data)
            for field_name, field_data in
            fields.items()
        ])

        namespace = f"""export declare module {self.get_namespace_name()} {{
{' ' * (INDENT * self.indent)}{fields_interfaces}
{' ' * (INDENT * (self.indent - 1))}}}"""
        return namespace

    def get_namespace_name(self):
        serializer_name = Transpiler(self.serializer).get_serializer_name()
        return f"I{serializer_name}Schema"



import importlib
from pathlib import Path

from django.conf import settings
from django.core.management import BaseCommand
from rest_framework.serializers import BaseSerializer, SerializerMetaclass

from rest_serializer_to_typescript import Transpiler

DEFAULT_SETTINGS = {
    "META_DATA": "rest_framework.metadata.SimpleMetadata",
    "SERIALIZERS": {}
}


class Command(BaseCommand):
    def handle(self, *args, **options):
        final_settings = {
            **DEFAULT_SETTINGS,
            **getattr(settings, "REST_SERIALIZERS_TO_TYPESCRIPT", {})
        }

        target_serializers: dict[str, str] = final_settings.get("SERIALIZERS")
        if not target_serializers:
            return

        for serializer_module, export_target in target_serializers.items():
            print(f"Exporting {serializer_module} to {export_target}: Started")

            target_path = export_target.split("/")
            path_without_file = target_path[:-1]

            Path("/".join(path_without_file)).mkdir(parents=True, exist_ok=True)

            modules = importlib.import_module(serializer_module)

            serializers = list(
                filter(
                    lambda serializer_class: issubclass(serializer_class, BaseSerializer),
                    filter(
                        lambda unknown: type(unknown) == SerializerMetaclass,
                        map(
                            lambda attr: getattr(modules, attr),
                            dir(modules)
                        )
                    )
                )
            )

            typescript = ""
            for serializer in serializers:
                typescript += f"{Transpiler(serializer()).generate_ts()}\n\n"

            with open(export_target, "w") as file:
                file.write(typescript)

            print(f"Exporting {serializer_module} to {export_target}: Finished")



from django.conf import settings

settings.configure()

from rest_framework import serializers

from rest_serializer_to_typescript import Transpiler


def test_user_serializer():
    class UserSerializer(serializers.Serializer):
        first_name = serializers.CharField()
        last_name = serializers.CharField(required=False)
        age = serializers.IntegerField()
        birth_year = serializers.IntegerField(required=False)

    expected_result = """export interface IUserSerializer {
    __is_modal: false;
    first_name: string;
    last_name?: string;
    age: number;
    birth_year?: number;
}"""

    assert Transpiler(UserSerializer()).generate_ts() == expected_result


def test_users_serializer():
    class UserSerializer(serializers.Serializer):
        first_name = serializers.CharField()
        last_name = serializers.CharField(required=False)
        age = serializers.IntegerField()
        birth_year = serializers.IntegerField(required=False)

    class UsersSerializer(serializers.Serializer):
        users = UserSerializer(many=True)

    expected_result = """export interface IUsersSerializer__UserSerializer {
    __is_modal: false;
    first_name: string;
    last_name?: string;
    age: number;
    birth_year?: number;
}

export interface IUsersSerializer {
    __is_modal: false;
    users: IUsersSerializer__UserSerializer[];
}"""

    assert Transpiler(UsersSerializer()).generate_ts() == expected_result


def test_get_ts():
    class FooSerializer(serializers.Serializer):
        some_field = serializers.ListField(child=serializers.IntegerField())
        another_field = serializers.CharField()
        null_field = serializers.CharField(allow_null=True)

    class BarSerializer(serializers.Serializer):
        foo = FooSerializer()
        foos = FooSerializer(many=True)
        bar_field = serializers.CharField()

    expected = """export interface IBarSerializer__FooSerializer {
    __is_modal: false;
    some_field: number[];
    another_field: string;
    null_field: string | null;
}

export interface IBarSerializer {
    __is_modal: false;
    foo: IBarSerializer__FooSerializer;
    foos: IBarSerializer__FooSerializer[];
    bar_field: string;
}"""

    assert Transpiler(BarSerializer()).generate_ts() == expected


def test_serializer_method():
    class CustomSerializer(serializers.Serializer):
        value_1 = serializers.SerializerMethodField()
        value_2 = serializers.SerializerMethodField()
        value_3 = serializers.SerializerMethodField()
        value_4 = serializers.SerializerMethodField()

        def get_value_2(self) -> str:
            ...

        def get_value_3(self) -> int:
            ...

        def get_value_4(self):
            ...

    expected = """export interface ICustomSerializer {
    __is_modal: false;
    value_1?: unknown;
    value_2?: string;
    value_3?: number;
    value_4?: unknown;
}"""

    assert Transpiler(CustomSerializer()).generate_ts() == expected



from rest_framework.metadata import SimpleMetadata
from rest_framework.schemas.openapi import AutoSchema


class APIMetadata(SimpleMetadata):
    """Extended metadata generator."""

    def get_field_info(self, field):
        field_info = super().get_field_info(field)

        # Add extra validators using the OpenAPI schema generator
        validators = {}
        AutoSchema()._map_field_validators(field, validators)
        extra_validators = ["format", "pattern"]
        for validator in extra_validators:
            if validators.get(validator, None):
                field_info[validator] = validators[validator]

        field_info["initial"] = field.initial
        field_info["field_name"] = field.field_name
        field_info["write_only"] = field.write_only

        return field_info



from django.conf import settings

settings.configure()

from rest_framework import serializers

from rest_serializer_to_typescript import SerializerMetadataTypeGenerator
from rest_serializer_to_typescript.metadata import APIMetadata


def test_serializer_schema_with_single_string_field():
    class MySerializer(serializers.Serializer):
        some_field = serializers.CharField()

    metadata = APIMetadata()

    expected = """export declare module IMySerializerSchema {
    export interface some_field {
        type: "string";
        required: boolean;
        read_only: boolean;
        label: string;
        initial: string;
        field_name: "some_field";
        write_only: boolean;
    }
}"""

    assert SerializerMetadataTypeGenerator(MySerializer(), metadata=metadata).generate_schema() == expected


def test_user_serializer():
    metadata = APIMetadata()

    class UserSerializer(serializers.Serializer):
        first_name = serializers.CharField()
        last_name = serializers.CharField(required=False)
        age = serializers.IntegerField()
        birth_year = serializers.IntegerField(required=False)

    expected_result = """export declare module IUserSerializerSchema {
    export interface first_name {
        type: "string";
        required: boolean;
        read_only: boolean;
        label: string;
        initial: string;
        field_name: "first_name";
        write_only: boolean;
    }

    export interface last_name {
        type: "string";
        required: boolean;
        read_only: boolean;
        label: string;
        initial: string;
        field_name: "last_name";
        write_only: boolean;
    }

    export interface age {
        type: "integer";
        required: boolean;
        read_only: boolean;
        label: string;
        initial: unknown;
        field_name: "age";
        write_only: boolean;
    }

    export interface birth_year {
        type: "integer";
        required: boolean;
        read_only: boolean;
        label: string;
        initial: unknown;
        field_name: "birth_year";
        write_only: boolean;
    }
}"""

    assert SerializerMetadataTypeGenerator(UserSerializer(), metadata=metadata).generate_schema() == expected_result


def test_users_serializer():
    metadata = APIMetadata()

    class UserSerializer(serializers.Serializer):
        first_name = serializers.CharField()
        last_name = serializers.CharField(required=False)
        age = serializers.IntegerField()
        birth_year = serializers.IntegerField(required=False)

    class UsersSerializer(serializers.Serializer):
        users = UserSerializer(many=True)

    expected_result = """export declare module IUsersSerializerSchema {
    export declare module IUserSerializerSchema {
        export interface first_name {
            type: "string";
            required: boolean;
            read_only: boolean;
            label: string;
            initial: string;
            field_name: "first_name";
            write_only: boolean;
        }

        export interface last_name {
            type: "string";
            required: boolean;
            read_only: boolean;
            label: string;
            initial: string;
            field_name: "last_name";
            write_only: boolean;
        }

        export interface age {
            type: "integer";
            required: boolean;
            read_only: boolean;
            label: string;
            initial: unknown;
            field_name: "age";
            write_only: boolean;
        }

        export interface birth_year {
            type: "integer";
            required: boolean;
            read_only: boolean;
            label: string;
            initial: unknown;
            field_name: "birth_year";
            write_only: boolean;
        }
    }

    export interface users {
        type: "field";
        required: boolean;
        read_only: boolean;
        label: string;
        child: IUserSerializerSchema[];
        initial: unknown;
        field_name: "users";
        write_only: boolean;
    }
}"""

    assert SerializerMetadataTypeGenerator(UsersSerializer(), metadata=metadata).generate_schema() == expected_result

#
# def test_get_ts():
#     class FooSerializer(serializers.Serializer):
#         some_field = serializers.ListField(child=serializers.IntegerField())
#         another_field = serializers.CharField()
#         null_field = serializers.CharField(allow_null=True)
#
#     class BarSerializer(serializers.Serializer):
#         foo = FooSerializer()
#         foos = FooSerializer(many=True)
#         bar_field = serializers.CharField()
#
#     expected = """export interface IBarSerializer__FooSerializer {
#     __is_modal: false;
#     some_field: number[];
#     another_field: string;
#     null_field: string | null;
# }
#
# export interface IBarSerializer {
#     __is_modal: false;
#     foo: IBarSerializer__FooSerializer;
#     foos: IBarSerializer__FooSerializer[];
#     bar_field: string;
# }"""
#
#     assert Transpiler(BarSerializer()).generate_ts() == expected
#
#
# def test_serializer_method():
#     class CustomSerializer(serializers.Serializer):
#         value_1 = serializers.SerializerMethodField()
#         value_2 = serializers.SerializerMethodField()
#         value_3 = serializers.SerializerMethodField()
#         value_4 = serializers.SerializerMethodField()
#
#         def get_value_2(self) -> str:
#             ...
#
#         def get_value_3(self) -> int:
#             ...
#
#         def get_value_4(self):
#             ...
#
#     expected = """export interface ICustomSerializer {
#     __is_modal: false;
#     value_1?: unknown;
#     value_2?: string;
#     value_3?: number;
#     value_4?: unknown;
# }"""
#
#     assert Transpiler(CustomSerializer()).generate_ts() == expected
